"""Arcos4py top level module.

This package is a python port of an R package
to detect and analyse collective phenomena.
"""

__author__ = """Benjamin Graedel"""
__email__ = "benjamin.graedel@unibe.ch"
__version__ = '0.1.5'

from .arcos4py import ARCOS

__all__ = ["ARCOS"]
